##################################################
## Description: Companion code of research paper "Deep learning for time series classification in ecology" by César Capinha, Ana Ceia-Hasse, Andrew M. Kramer and Christiaan Meijer. This script implements data importation and modelling procedures for case study #1 described in the paper ("species identification")
## Author: Cesar Capinha
## Credits: Suggestions made by Ana Ceia-Hasse, Andrew M. Kramer & Christiaan Meijer helped improve the script. Parts of the script are adapted from code made available by the 'mcfly' team (https://github.com/NLeSC/mcfly)
## Email: cesarcapinha at campus.ul.pt
## Software versions: Python 3.7.1, running on Anaconda; Tensorflow v2.0.0; mcfly v.3.0
##################################################

############################
#Load csvs into numpy arrays
import os
import numpy as np

os.chdir("C:/Deep_learning_ecology/Data/Case_study_1")

#########
#Load At#
#########
At_X_2dim = np.loadtxt(open("At_X.csv", "rb"), delimiter=",", skiprows=1)

# In this case, the X data has only one dimension ('channel').
# We need to structure it as a tensor
# this does not alter the data but allows mcfly (and keras) to read the data.
At_X = np.expand_dims(At_X_2dim, axis=2)

#Load At_Y data 
At_Y = np.loadtxt(open("At_Y.csv", "rb"), delimiter=",", skiprows=1)


#########
#Load Av#
#########
Av_X_2dim = np.loadtxt(open("Av_X.csv", "rb"), delimiter=",", skiprows=1)
Av_X = np.expand_dims(Av_X_2dim, axis=2)

#Load Av_Y data 
Av_Y = np.loadtxt(open("Av_Y.csv", "rb"), delimiter=",", skiprows=1)


#########
#Load Bt#
#########
Bt_X_2dim = np.loadtxt(open("Bt_X.csv", "rb"), delimiter=",", skiprows=1)
Bt_X = np.expand_dims(Bt_X_2dim, axis=2)

#Load Bt_Y data 
Bt_Y = np.loadtxt(open("Bt_Y.csv", "rb"), delimiter=",", skiprows=1)


#########
#Load Bv#
######### 
Bv_X_2dim = np.loadtxt(open("Bv_X.csv", "rb"), delimiter=",", skiprows=1)
Bv_X = np.expand_dims(Bv_X_2dim, axis=2)

#Load Bv_Y data 
Bv_Y = np.loadtxt(open("Bv_Y.csv", "rb"), delimiter=",", skiprows=1)


#########
#Load T #
######### 
T_X_2dim = np.loadtxt(open("T_X.csv", "rb"), delimiter=",", skiprows=1)
T_X = np.expand_dims(T_X_2dim, axis=2)

#Load T_Y data 
T_Y = np.loadtxt(open("T_Y.csv", "rb"), delimiter=",", skiprows=1)
####END OF LOAD DATA####


###########
#Run McFly#
###########
import sys
import os
import datetime
import pandas as pd
import tensorflow as tf
# mcfly
from mcfly import modelgen, find_architecture, storage
from keras.models import load_model
from time import process_time

#set random seed for replicability
np.random.seed(0)
tf.random.set_seed(0)


# Specify in which directory you want to store the data:
directory_to_extract_to = 'C:/Deep_learning_ecology/Results'
folder= 'Case_study_1'


#Inspect data
#see the dimensions of At_X numpy arrays- change as required
print('x shape:', At_X.shape)
print('y shape:', At_Y.shape)

#see sizes
print('At set size:', At_X.shape[0])


#Define class labels
labels = ['bole', 'amel', 'laris']
# bole corresponds to Bactrocera oleae, amel to Apis mellifera, laris to Lonchaea aristella


#####################################################################
#Start procedures for generation of distinct modelling architectures#
####################################################################

#identify number of classes
num_classes = At_Y.shape[1]

#At code below - 'number_of_models'
#Because mcfly uses random search, you will get better results when using more models.
models = modelgen.generate_models(At_X.shape,
                                  number_of_classes=num_classes,
                                  number_of_models = 20)



#Print generated architectures
models_to_print = range(len(models))
for i, item in enumerate(models):
    if i in models_to_print:
        model, params, model_types = item
        print("-------------------------------------------------------------------------------------------------------")
        print("Model " + str(i))
        print(" ")
        print("Hyperparameters:")
        print(params)
        print(" ")
        print("Model description:")
        model.summary()
        print(" ")
        print("Model type:")
        print(model_types)
        print(" ")
        
         
# Set directory where the results, e.g. json file, will be stored
resultpath = os.path.join(directory_to_extract_to, folder)
if not os.path.exists(resultpath):
        os.makedirs(resultpath)


#############################################
#Start training of preliminary architectures#
#############################################

#We can also adjust the number of times the subset is iterated over ('nr_epochs') and opt to use a speciefied subset of instances in the train partial data sets ('subset_size').
outputfile = os.path.join(resultpath, 'modelcomparison.json')

start_time = datetime.datetime.now() 

histories, val_accuracies, val_losses = find_architecture.train_models_on_samples(At_X, At_Y,
                                                                           Av_X, Av_Y,
                                                                           models, nr_epochs=4,
                                                                           subset_size=At_X.shape[0],
                                                                           verbose=True,
                                                                           outputfile=outputfile)
print('Details of the training process were stored in ',outputfile)

#Save summary of model comparisons
#Note: ['acc'] and ['val_acc']' were changed to ['accuracy'] and ['val_accuracy'], respectively for compatibility with current mcfly version
modelcomparisons = pd.DataFrame({'model':[str(params) for model, params, model_types in models],
                       'train_acc': [history.history['accuracy'][-1] for history in histories],
                      'train_loss': [history.history['loss'][-1] for history in histories],
                       'val_acc': [history.history['val_accuracy'][-1] for history in histories],
                      'val_loss': [history.history['val_loss'][-1] for history in histories]
                     })
modelcomparisons.to_csv(os.path.join(resultpath, 'modelcomparisons.csv'))
    

#Select best performing candidate model
best_model_index = np.argmax(val_accuracies)
best_model, best_params, best_model_types = models[best_model_index]
print('Model type and parameters of the best model:')
print(best_model_types)
print(best_params)


#Train of selected architecture on the full train data
#Repeat this section for one training step of n epochs
nr_epochs = 72
datasize = Bt_X.shape[0] # Change in `X_train.shape[0]` if training complete data set
modelname = 'trainedmodel.h5'

#train during x epochs
for x in range(nr_epochs):
	history = best_model.fit(Bt_X[:datasize,:,:], Bt_Y[:datasize,:], epochs=1, verbose=True)
	fname=str(x)
	model_path = os.path.join(resultpath,fname)
	if not os.path.exists(model_path): os.makedirs(model_path)
	os.chdir(model_path)
	best_model.save('trainedmodel.h5')
	y_pred = best_model.predict(Bv_X)
	np.savetxt("PREDICTIONS_bole.csv",  y_pred[:,0], delimiter=",")
	np.savetxt("TRUE_VALUES_bole.csv", Bv_Y[:,0], delimiter=",")
	np.savetxt("PREDICTIONS_amel.csv",  y_pred[:,1], delimiter=",")
	np.savetxt("TRUE_VALUES_amel.csv", Bv_Y[:,1], delimiter=",")
	np.savetxt("PREDICTIONS_laris.csv",  y_pred[:,2], delimiter=",")
	np.savetxt("TRUE_VALUES_laris.csv", Bv_Y[:,2], delimiter=",")    
	print(x)
#End of training step section

end_time = datetime.datetime.now()


#######################################
#Load final model for  test prediction#
#######################################
#!Specify which directory holds the final model (note that in python the indexing starts from 0):
model_path = 'C:/Deep_learning_ecology/Results/Case_study_1/46/trainedmodel.h5'
model_reloaded = tf.keras.models.load_model(model_path)
